signature CANON =  sig
  include PASS
  datatype canon =
    Ws  of (Context.ctxt * string)
  | Txt of (Context.ctxt * string)
  | Seq of canon * canon

  val canon   : SMLNG.doc -> canon
  val toDoc   : canon -> SMLNG.doc 
  val toList  : SMLNG.doc -> (Context.ctxt * string * bool) list
  val compareDoc : SMLNG.doc * SMLNG.doc -> (int * int * bool) 

end
structure Canon :> CANON =
  struct
  datatype canon = 
    Txt of (Context.ctxt * string)
  | Ws  of (Context.ctxt * string)
  | Seq of canon * canon
    
  fun canon doc =
    Context.fold_with_ctxt {kSeq=Seq,
			    kTxt=Txt,
			    kWs=Ws,
			    kNd=(fn (_,t,a) => a)} Context.init doc

  fun mkNode f [] t = f t
    | mkNode f (x::xs) t = SMLNG.Nd(x,mkNode f xs t)

  fun toDoc (Txt(ctxt,text)) = mkNode SMLNG.Txt (Context.toTags ctxt) text
    | toDoc (Ws(ctxt,text)) = mkNode SMLNG.Ws (Context.toTags ctxt) text
    | toDoc (Seq(d1,d2)) = SMLNG.Seq(toDoc d1,toDoc d2)

    fun tagsize SMLNG.EM = 9
      | tagsize SMLNG.PL = 9
      | tagsize SMLNG.TT = 9
      | tagsize _ = 7

  fun isWs c = Context.Get.u c = 0w0 andalso not (Context.Get.tt c)
  fun sz_list doc =
    Context.fold_with_ctxt
    {kSeq=(fn ((sx,x),(sy,y)) => (sx + sy,JoinList.append(x,y))),
     kTxt=(fn (c,s) => (String.size s,JoinList.one (c,s,false))),
     kWs=(fn (c,s) => (String.size s,JoinList.one (c,s,isWs c))),
     kNd=(fn (_,t,(sa,a)) => ((tagsize t) + sa,a))} Context.init doc

  fun toList x = let
    val (_,d) = sz_list x
  in JoinList.toList d
  end
  fun compareDoc (d1,d2) = let
    val (s1,d1) = sz_list d1
    val (s2,d2) = sz_list d2
    fun eat_ws s = let
      fun loop (JoinList.Nil) = JoinList.Nil
	| loop (JoinList.Cons((_,"",_),xs)) = loop (JoinList.pat xs)
	| loop (JoinList.Cons((c,_,true),xs)) =	let
	  val next = JoinList.pat xs
	  in  case next of
	    JoinList.Nil => next
	  | JoinList.Cons((_,"",_),ys) => loop(JoinList.pat ys)
	  | JoinList.Cons((c',_,true),ys) => 
	      if Context.eq(c',c) then loop(JoinList.pat ys)
	      else next
	  |  _ => next
	  end
	| loop x = x
    in loop (JoinList.pat s)
    end
    fun eq (JoinList.Nil,JoinList.Nil) = true
      | eq (JoinList.Cons((cx,sx,wx),xs),JoinList.Cons((cy,sy,wy),ys)) =
       wx = wy andalso Context.eq (cx,cy) andalso 
       (sx = sy) andalso eq(eat_ws xs,eat_ws ys)
      | eq (xs,ys) = false
  in  (s1,s2,eq(eat_ws d1,eat_ws d2))
  end
    val name = "canon"
    val go = toDoc o canon
    fun tests () = ()
  end