signature JOIN_LIST =
  sig
    datatype ('a,'b) T =
      Nil 
    | Cons of 'a * 'b
    type 'a jlist

    val pat      : 'a jlist -> ('a,'a jlist) T
    val con      : ('a,'a jlist) T -> 'a jlist 

    val fold     : (('a,'b) T -> 'b) -> 'a jlist -> 'b
    val unfold   : ('b -> ('a,'b) T) -> 'b -> 'a jlist

    val empty    : 'a jlist
    val one      : 'a -> 'a jlist
    val append   : ('a jlist * 'a jlist) -> 'a jlist

    val isEmpty  : 'a jlist -> bool
    val length   : 'a jlist -> int

    val toList   : 'a jlist -> 'a list
    val fromList : 'a list -> 'a jlist
  end

structure JoinList :> JOIN_LIST =
  struct
    datatype ('a,'b) T =
      Nil 
    | Cons of 'a * 'b

    datatype 'a jlist =
      Empty
    | One of 'a
    | App of 'a jlist * 'a jlist
    val empty = Empty
    val one = One
    fun jCons(x,xs) = append(One x,xs)
    and append (xs,Empty) = xs
      | append (Empty,ys) = ys
      | append (xs,ys) = App(xs,ys)

    fun pat Empty = Nil
      | pat (One x) = Cons(x,Empty)
      | pat (App(xs,ys)) =
      case (pat xs) of
	Cons(x,xs) => Cons(x,append(xs,ys))
      | _ => raise Fail "impossible"

    fun con Nil = Empty
      | con (Cons (x,xs)) = jCons(x,xs)

    fun fold f Empty = f Nil
      | fold f (One x) = f (Cons(x,f Nil))
      | fold f (App(xs,ys)) =
      let
	val b = fold f ys
	fun rest Nil = b
	  | rest (Cons(x,xs)) = f (Cons(x,xs))
      in fold rest xs
      end

    fun isEmpty Empty = true
      | isEmpty _ = false
    fun unfold g x =
      case (g x) of
	Nil => Empty
      | Cons(x,xs) => jCons(x,unfold g xs)

    fun length xs =
      let
	fun loop ([],acc) = acc
	  | loop (Empty::xs,acc) = loop(xs,acc)
	  | loop ((One _)::xs,acc) = loop(xs,acc+1)
	  | loop (App(xs,ys)::rest,acc) = loop(xs::ys::rest,acc)
      in loop([xs],0)
      end
    fun toList x = fold (fn Nil => [] | Cons(x,xs) => x::xs) x
    fun fromList x = unfold (fn [] => Nil | (x::xs) => Cons(x,xs) ) x
  end
