structure Main =
  struct
    val {dbg,dsay,derr,ddoc} = Debug.mk ()

    fun parse s = let
      val ins = TextIO.openIn s
      val doc = SMLNG.parse ins
    in TextIO.closeIn ins;doc
    end

    fun check n f old = let
      val new = f old
      val (new_sz,old_sz,same) = Canon.compareDoc (new,old)
      val bigger = new_sz > old_sz
      val _ = dsay ("\n# pass: "^n);
    in (if not same then raise (derr "\n# Not Same")
	else if bigger then raise (derr "\n# Bigger")
        else (dsay ("\n# change: "^(Int.toString (old_sz - new_sz)));
	      new))
      handle e => (dsay ("\n# FAIL Recovering");
		   dsay ("\n# input: "^n); ddoc old;
		   dsay ("\n# output: "^n); ddoc new; raise e)
    end
    local
      val last_good = ref (SMLNG.Txt "")
    in
      fun do_pass n f x = let
	val _ = last_good := x
	val x = check n f x
      in last_good := x;x
      end
      fun try_pass n f x = let
	val _ = last_good := x
	val x = (check n f x) handle _ => x
      in last_good := x;x
      end
      fun fail_safe t f x =
	((TimeLimit.timeLimit (Time.fromSeconds t) f x) handle _ =>
	  (dsay("\n# Fail safe invoked"); !last_good))
    end
    structure CA = Canon
    structure MA = MergeAdjacent
    structure RD = RemoveRedundant
    structure EME = EMElimination
    structure OEME = OldEMElimination
    structure CO = ColorOpt
    structure SO = SizeOpt
    structure WS = WSSimplification 
    structure ING = InterNG

    fun merge_opts opts = let
      val name = List.foldr (fn ((n,_),s) => n^"-"^s) "" opts
      val f = List.foldr (fn ((n,f),g) => (g o f)) (fn x => x) opts
    in try_pass name f
    end

    fun tests () =
      List.app (fn f => f ())
      [CA.tests,
       RD.tests,
       EME.tests,
       MA.tests,
       WS.tests,
       ING.tests]

    fun loop x = loop x
    val opts =
      [(RD.name,RD.go),
       (MA.name,MA.go),
       (EME.name, EME.go),
       (WS.name,  WS.go), 
       (ING.name, ING.go),
       (CO.name, CO.go),
       (SO.name, SO.go)]


      (* try each pass check after each phase for correctness *)
    val safe_opts = 
      List.foldr (fn ((n,f),g) => (g o (try_pass n f))) (fn x => x) opts

    val clean_opts =
      [(RD.name,RD.go),
       (MA.name,MA.go),
       (CO.name, CO.go),
       (SO.name, SO.go),
       (WS.name, WS.go)]

    val opt = (merge_opts clean_opts) o safe_opts

    fun toInt dflt s = Option.getOpt(LargeInt.fromString s,dflt)
    val dflt_tlimit = LargeInt.fromInt 180

    fun main (_,args) = let
      val (tlimit,doc) =
	case args of [x] =>
	  (toInt dflt_tlimit x,SMLNG.parse  TextIO.stdIn)
      | (x::y::_) => (toInt dflt_tlimit x,parse y)
      | [] => (dflt_tlimit,SMLNG.parse  TextIO.stdIn)
      val slop = 5 (* seconds of slack time *)
      val doc = fail_safe (tlimit - slop) opt doc
    in dsay "\n# Answer \n";
      SMLNG.print doc TextIO.stdOut;
      OS.Process.success
    end
  end
