signature REMOVE_REDUNDANT = sig
  include PASS
  val dbg : bool ref
end
structure RemoveRedundant :> REMOVE_REDUNDANT =
  struct
    structure T = SMLNG
    structure C = Context
    val {dsay,derr,ddoc,dbg} = Debug.mk ()
    val name = "remove-redundant"
    fun go doc = let
      fun kSeq(x,y) = T.Seq(x,y)
      fun kTxt(c,s) = T.Txt s
      fun kWs(c,s) = T.Ws s
      fun kNd (c,t,T.Txt "") = T.Txt ""
	| kNd (c,t,T.Ws "") = T.Ws ""
	| kNd (c,t,d) = if Context.isNop(t,c) then d
			else T.Nd(t,d)
      val doc = Context.fold_with_ctxt 
	{kSeq=kSeq, kTxt=kTxt, kWs=kWs, kNd=kNd} Context.init doc 
    in doc
    end
    val tests = Tests.mk name false go
      ["<b><r>1</r>2</b>",
       "<b><b>1</b>2</b>",
       "<EM><EM>1</EM></EM>",
       "<S><EM><EM>1</EM></EM></S>",
       "<U><U><U><U>1</U></U></U></U>",
       "<S><EM><EM>1</EM></EM></EM></S>",
       "<S><EM><EM>1<U><U><U><U>2</U></U></U></U>"^
       "<b><b>1</b></b></EM></EM></EM></S>"
       ]
  
  end
